from dune.perftool.generation import (include_file,
                                      preamble,
                                      )
from dune.perftool.pdelab.driver import (get_formdata,
                                         get_trial_element,
                                         name_initree,
                                         preprocess_leaf_data,
                                         )
from dune.perftool.pdelab.driver.gridfunctionspace import (name_leafview,
                                                           name_gfs,
                                                           name_trial_gfs,
                                                           type_leafview,
                                                           )
from dune.perftool.pdelab.driver.solve import name_vector


@preamble
def define_vtkfile(name):
    ini = name_initree()
    include_file("string", filetag="driver")
    return "std::string {} = {}.get<std::string>(\"wrapper.vtkcompare.name\", \"output\");".format(name, ini)


def name_vtkfile():
    define_vtkfile("vtkfile")
    return "vtkfile"


@preamble
def typedef_vtkwriter(name):
    include_file("dune/grid/io/file/vtk/subsamplingvtkwriter.hh", filetag="driver")
    gv = type_leafview()
    return "using {} = Dune::SubsamplingVTKWriter<{}>;".format(name, gv)


def type_vtkwriter():
    typedef_vtkwriter("VTKWriter")
    return "VTKWriter"


@preamble
def define_subsamplinglevel(name):
    ini = name_initree()
    return "int {} = {}.get<int>(\"vtk.subsamplinglevel\", 0);".format(name, ini)


def name_subsamplinglevel():
    define_subsamplinglevel("sublevel")
    return "sublevel"


@preamble
def define_vtkwriter(name):
    _type = type_vtkwriter()
    gv = name_leafview()
    subsamp = name_subsamplinglevel()
    return "{} {}({}, {});".format(_type, name, gv, subsamp)


def name_vtkwriter():
    define_vtkwriter("vtkwriter")
    return "vtkwriter"


@preamble
def vtkoutput():
    include_file("dune/pdelab/gridfunctionspace/vtk.hh", filetag="driver")
    vtkwriter = name_vtkwriter()
    gfs = name_trial_gfs()
    define_gfs_names(get_trial_element())
    vtkfile = name_vtkfile()
    predicate = name_predicate()
    vec = name_vector(get_formdata())

    return ["Dune::PDELab::addSolutionToVTKWriter({}, {}, {}, Dune::PDELab::vtk::defaultNameScheme(), {});".format(vtkwriter, gfs, vec, predicate),
            "{}.write({}, Dune::VTK::ascii);".format(vtkwriter, vtkfile)]


def type_predicate():
    include_file("dune/perftool/vtkpredicate.hh", filetag="driver")
    return "CuttingPredicate"


@preamble
def define_predicate(name):
    t = type_predicate()
    return "{} {};".format(t, name)


def name_predicate():
    define_predicate("predicate")
    return "predicate"



@preamble
def typedef_vtk_sequence_writer(name):
    include_file("dune/grid/io/file/vtk/vtksequencewriter.hh", filetag="driver")
    gv_type = type_leafview()
    return "using {} = Dune::VTKSequenceWriter<{}>;".format(name, gv_type)


def type_vtk_sequence_writer():
    typedef_vtk_sequence_writer("VTKSW")
    return "VTKSW"


@preamble
def define_vtk_sequence_writer(name):
    vtksw_type = type_vtk_sequence_writer()
    vtkw_type = type_vtkwriter()
    vtkw = name_vtkwriter()
    vtkfile = name_vtkfile()
    return "{} {}(std::make_shared<{}>({}), {});".format(vtksw_type, name, vtkw_type, vtkw, vtkfile)


def name_vtk_sequence_writer():
    define_vtk_sequence_writer("vtkSequenceWriter")
    return "vtkSequenceWriter"


@preamble
def visualize_initial_condition():
    include_file("dune/pdelab/gridfunctionspace/vtk.hh", filetag="driver")
    vtkwriter = name_vtk_sequence_writer()
    element = get_trial_element()
    define_gfs_names(element)
    gfs = name_trial_gfs()
    vector = name_vector(get_formdata())
    predicate = name_predicate()
    from dune.perftool.pdelab.driver.instationary import name_time
    time = name_time()
    return ["Dune::PDELab::addSolutionToVTKWriter({}, {}, {}, Dune::PDELab::vtk::defaultNameScheme(), {});".format(vtkwriter, gfs, vector, predicate),
            "{}.write({}, Dune::VTK::appendedraw);".format(vtkwriter, time)]


@preamble
def _define_gfs_name(element, is_dirichlet, offset):
    from ufl import MixedElement, VectorElement
    if isinstance(element, VectorElement):
        gfs = name_gfs(element, is_dirichlet)
        return "{}.name(\"data_{}to{}\");".format(gfs, offset, offset + element.value_size())
    elif isinstance(element, MixedElement):
        k = 0
        for subel in element.sub_elements():
            _define_gfs_name(subel, is_dirichlet[k:k + subel.value_size()], offset + k)
            k = k + subel.value_size()
        return []
    else:
        # This is a scalar leaf
        gfs = name_gfs(element, is_dirichlet)
        return "{}.name(\"data_{}\");".format(gfs, offset)


def define_gfs_names(element):
    is_dirichlet = preprocess_leaf_data(element, "is_dirichlet")
    return _define_gfs_name(element, is_dirichlet, 0)